import os
import io
import re
from functools import wraps
from .frontend.display import Print,Verbose
from .urls import Urls
from .player import Player
from .errors import MediaError
from .utils.request import Session
from .backend.filehandler import file_size,Tmp,Path
from .backend.mediasetup import Album,Mp3
from .backend.config import User,Datatype,Queued
import traceback


class Media():
    """ Media player that control the songs selected from Mixtapes """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_tmpfile'):
            Tmp.removeTmpOnstart()
            cls._tmpfile = Tmp.create()
        return super(Media, cls).__new__(cls)


    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.mixtape)


    def __repr__(self):
        return 'Media(%s)' % (self.mixtape)


    def __len__(self):
        if hasattr(self,'songs'):
            return len(self.songs)
        else:
            return 0


    def __init__(self, mixtape=None):
        """ Initialize Media 

        @@params: mixtape - Datpiff.Mixtapes object
        """
        if 'mixtapes.Mixtapes' not in str(type(mixtape)):
            raise MediaError(2,'must pass a mixtape object to Media class')

        Verbose('Media initialized')
        if not mixtape:
            raise MediaError(1)

        self._session = Session()
        self.mixtape = mixtape
        self._artist_name = None
        self.album_name = None
        self._current_index = None
        self._selected_song = None
        self._downloaded_song = None
        super(Media, self).__init__()

    
    def findSong(self,songname):
        songname = Datatype.strip_lowered(songname)
        links = self.mixtape.links
        links = list(enumerate(links,start=1))
        results = Queued(self._search,links,songname).run()
        if not results:
            Print('No song was found with this title ')
        results = Datatype.removeNone(results)
        return results
        """ 
        links = lists(enumerate(links,start=1))
        collector = []
        for index, link in enumerate(links,start=1):
            album = Album(link)
            name = album.name
            tracks = Mp3(album.embed_response).songs
            if any(songname in Datatype.strip_lowered(track) for track in tracks):
                collector.append((index,name)) 

        return collector
        """
        
    def _search(self,links,song):
        index,link = links
        album = Album(link)
        name = album.name
        tracks = Mp3(album.embed_response).songs
        if any(song in Datatype.strip_lowered(track) for track in tracks):
            return index,name 

       
        

    def setMedia(self, selection):
        """
        Setup and queues Datpiff.Mixtapes ablum to media player by selection 
        Datpiff.Mixtapes artists or album name.

        @@params: selection - Datpiff.Mixtapes album's name or artist's name.
            #datatype 
            int - will return the Datpiff.Mixtape artist at that index.
            str - will search for an artist from Mixtapes.artists (default)
                  or album from Mixtapes.ablum. 
        """
        result = self.mixtape._select(selection)
        if result is None:
            Verbose('SELECTION:',selection)
            e_msg = '\n--> Mixtape "%s" was not found'%selection
            raise MediaError(1)

        self._artist_name = self.mixtape.artists[result]
        self.album_name = self.mixtape.mixtapes[result]
        link = self.mixtape._links
        self._setup(link[result])
        
        Verbose('Setting Media to %s - %s' % (self.artist, self.album))
        # only returning to check if choice was set
        #return choice


    def _setup(self,link):
        self._album = Album(link)
        response = self._album.embed_response
        self._Mp3 = Mp3(response)
        self._song_cache = {}

       
    def _parseSelection(self, select):
        """Parse all user selection and return the correct songs
           @@params: select  - Media.songs name or index of Media.songs 
        """
        try:
            return User.selection(select,self.songs,[x.lower() for x in self.songs])
        except MediaError as e:
            raise MediaError(5)



    @property
    def artist(self):
        """Return the current artist name."""
        return self._artist_name


    @artist.setter
    def artist(self, name):
        """Set the current artist name."""
        self.setMedia(name)


    @property
    def album(self):
        """Return the current album name."""
        return self.album_name
        return Album.name


    @album.setter
    def album(self, name):
        self.setMedia(name)


    @property
    def songs(self):
        """ Return all album songs."""
        if not hasattr(self,'_Mp3'):
            e_msg = 'Set media by calling -->  Media.setMedia("Album name")'
            raise MediaError(3,e_msg)
        return self._Mp3.songs


    @property
    def mp3urls(self):
        """Returns the parsed mp3 url"""
        return list(self._Mp3.mp3Urls)


    @property
    def show_songs(self):
        """Pretty way to Print all song names"""
        try:
            songs = self.songs
            [Print('%s: %s' % (a+1, b)) for a, b in enumerate(songs)]
        except TypeError:
            Print("Please set Media first\nNo Artist name")


    @property
    def song(self):
        """Returns the current song set by user."""
        return self._selected_song


    @song.setter
    def song(self, name):
        """ 
        Set current song
        @@params: name  - Media.songs name or index of Media.songs 
        """
        songs = self.songs
        index = self._parseSelection(name)
        if index is not None:
            self._selected_song = songs[index]
            self._current_index = index
        else:
            Print('\n\t song was not found')


    def _cacheSong(self, song, data):
        """
        _cacheSong - Captures the songname and content when user play
         or download a song. This prevents downloading song content twice
         when playing or downloading a song. Data from each song will be 
         stored in _song_cache for future access.
        """
        name = "-".join((self.artist, song))
        try:
            self._song_cache[name] = data
        except MemoryError:
            self._song_cache = {}



    def _checkCache(self, song):
        """Check if song have already been download.
        If so,requests response is return.
        """
        in_cache = '-'.join((self.artist, song))
        if hasattr(self, '_song_cache'):
            if in_cache in self._song_cache:
                response = self._song_cache[in_cache]
                return response


    def mp3Content(self,track):
        """
        Return content of the song in IO Bytes object

        @@params: track - name or track index of song   
        """

        selection = self._parseSelection(track)
        if selection is None:
            return 

        self._song_index = selection
        link = self.mp3urls[selection]
        songname = self.songs[selection]
        self.song = selection + 1

        # Write songname to file
        # check if song has been already downloaded 
        # if so then get the response from cache
        response = self._checkCache(songname)
        if not response:
            response = self._session.method('GET', link)
            self._cacheSong(songname, response)

        return io.BytesIO(response.content)


    def play(self, track=None, demo=False):
        """ 
        Play song (uses vlc media player) 
         @@params: track - name or index of song type(str or int)
         @@params: demo  - True: demo sample of song
                              False: play full song 
                              *default: False
        """
        if track is None:
            Print('\n\t -- No song was entered --')
            return 

        if isinstance(track,int):
            if track > len(self):
                raise MediaError(4) 

        try:
            content = self.mp3Content(track).read()
        except Exception:
            Print('\n\t-- No song was found --')
            return 

        songname = self.songs[self._song_index]
        track_size = len(content)
        # play demo or full song
        if not demo:  # demo whole song
            chunk = content
            samp = int(track_size)
        else:  # demo partial song
            samp = int(track_size/5)
            start = int(samp/5)
            chunk = content[start:samp+start]

        with open(self._tmpfile.name, "wb") as ws:
            ws.write(chunk)
        sorf = 'Demo' if demo else 'Full Song'
        Verbose('\n%s %s %s' % ('-'*20, sorf, '-'*20))
        Verbose('Song: %s - %s' % (self.artist, songname))
        Verbose("Size:", file_size(samp))
        song = " - ".join((self.artist, songname))
        if not hasattr(self, 'player'):
            self.player = Player()
        
        self.player.setTrack(song,self._tmpfile.name)
        self.player.play


    def download(self, track=False, output="", name=None):
        """
        Download song from Datpiff
        @@params: track - name or index of song type(str or int)
        @@output: location to save the song (optional)
        @@name:   rename the song (optional)
            default will be song's name 
        """
        selection = self._parseSelection(track)
        if selection is None:
            return

        output = output or os.getcwd()
        if not Path.is_dir(output):
            Print('Invalid directory: %s'%output)
            return 

        link = self.mp3urls[selection]
        song = self.songs[selection]
        title  = ' - '.join(( self.artist,song.strip()+".mp3" ))
        title = Path.toStandard(title)
        songname = Path.join(output,title)

        try:
            response = self._checkCache(song)
            if response:
                content = response.content
            else:
                response = self._session.method('GET', link)
                response.raise_for_status()

            with open(songname, "wb") as ws:
                ws.write(response.content)
                Verbose('\nSONGNAME: ', songname,
                               '\nSIZE:  ', file_size(len(response.content)))
            self._cacheSong(songname, response)
        except:
            Print('Cannot download song %s' % songname)


    def downloadAlbum(self, output=None):
        if not output:
            output = os.getcwd()
        elif not os.path.isdir(output):
            Print('Invalid directory: %s'%output)
            return


        title = "-".join((self.artist, self.album))
        title = Path.toStandard(title)
        fname = Path.join(output,title) 
        
        # make a directory to store all the ablum's songs
        if not os.path.isdir(fname):
            os.mkdir(fname)

        for num, song in enumerate(self.songs):
            self.download(song, output=fname)
        Print("\n%s %s saved" % (self.artist, self.album))




"""
    This file will stored all dynamic class and methods.
    These methods will be used throughout the whole program. 
    
"""
import threading
from queue import Queue
from functools import wraps


class Queued():
    def __init__(self,main_job,input_work,search):
        self.input = input_work # work to put in queue
        self.main_job = main_job # job to perform with work
        # song to search for TODO: move this to main function
        self.search = search

        self.THREAD_COUNT = 75
        self.results = []
        self.q = Queue()

    def set_queue(self):
        if Datatype.isList(self.input):
            # setting the queue from ^
            for obj in self.input: # |
                self.q.put(obj)


    def put_worker_to_work(self):
        worker = self.q.get()
        self.results.append(self.main_job(worker,self.search))
        self.q.task_done()


    def start_thread(self):
        for _ in range(self.THREAD_COUNT):
             t = threading.Thread(target = self.put_worker_to_work)
             t.daemon = True
             t.start()

    
    def run(self):
        self.set_queue()
        while True:
            self.start_thread()
            if self.q.empty():
                break
        # data will not be filter here for 'None type' in list
        # must catch all None types in the base method 
        return self.results




class Datatype():
    @staticmethod 
    def isDict(_type):
        return isinstance(_type,dict)

    @staticmethod 
    def isStr(_type):
        return isinstance(_type,str)

    @staticmethod 
    def isList(_type):
        return isinstance(_type,list)

    @classmethod
    def removeNone(cls,_list):
        if not cls.isList(_list):
            msg = 'Can not remove None value. ..datatype must be a list '
            raise NotImplementedError(msg)
        return list(filter(None,_list))

    @classmethod
    def enumerate_it(cls,data,start=0):
        """Return enumerate object"""
        if cls.isDict(data) or cls.isList(data):
            if cls.isDict(data):
                data = data.items()
            return list(enumerate(data,start=start))
        raise NotImplementedError('datatype is not a dictionary or list ')


    @staticmethod 
    def strip_lowered(string):
        """Return strip and lower value"""
        return str(string).lower().strip()


    @classmethod
    def lowered_dict(cls,data):
        """Strip and lower keys in dictionary"""
        if not cls.isDict(data):
            raise NotImplementedError('datatype is not a dictionary')

        item = {}
        for key,val in data.items():
            item[cls.strip_lowered(key)] =  val
        return item


    @classmethod
    def lowered_list(cls,data):
        """Strip and lower string in list"""
        if not cls.isList(data):
            raise NotImplementedError('datatype is not a List')
        return [cls.strip_lowered(x) for x in data]


class User():
    @staticmethod
    def choice_is_str(choice,data):
        """
        Parse user string choice and return the corresponding data
        :params:: choice,data 
            choice - user choice. datatype: str
            data - list or dict object
        """
        choice =  Datatype.strip_lowered(choice)
        if  Datatype.isDict(data):
            data =  Datatype.lowered_dict(data)
            val = [val for key,val in data.items() if choice in key] 
        else:
            val = [val for val in data if choice in  Datatype.strip_lowered(val)] 

        if val:
            return min(val)


    @staticmethod
    def choice_is_int(choice,data):
        """
        Parse user int choice and return the corresponding data
        :params:: choice,data 
            choice - user choice. datatype: str
            data - list or dict object
        """

        try:
            choice = int(choice)
            results =  Datatype.enumerate_it(data)
            length = len(data)
            if  Datatype.isDict(data):
                return results[choice][1][1]
            return results[choice][1]
        except Exception as e:
            print(e)
            return 


    @classmethod
    def selection(cls,select,byint=None,bystr=None):
        ''' select ablums_link by artist name, album name ('title')
         or by index number of title or artist
        '''
        try:
            # checking from index 
            if isinstance(select,int):
                select-=1
                length = len(byint) + 1
                # catch index errors if user choose a mixtape out of range
                select = 0 if (0 >= select or select > len(byint)) else select
                return select 

            else:
                #checking from artists
                choosen =  cls.choice_is_str
                choice = choosen(select,byint)
                if choice: return byint.index(choice)
                
                # check by mixtapes
                choice = choosen(select,bystr)
                if choice: return bystr.index(choice)

                if not choice: # will catch None value in Media
                   return 
                
        except BuildError as e:
            raise BuildError(1)




"""
   Custom printing function using logger. 

   Stdout can be disable on startup by setting frontend verbose flag False.
   example:: frontend.verbose = True
"""
import logging
import pydatpiff 

rootLogger = logging.getLogger()
rootLogger.setLevel(logging.INFO)

def Print(*args):
    if args:
        args = "  ".join(args)
        rootLogger.info(args)

def Verbose(*args):
    if not pydatpiff.pydatpiff.verbose:
        return 
    Print(*args)





import re
from functools import wraps

ampersands = [
	"&quot;", "&amp;",
	"&lt;", "&gt;",
	"&nbsp;", "&iexcl;",
	"&cent;", "&pound;",
	"&curren;", "&yen;",
	"&brvbar;", "&sect;",
	"&uml;", "&copy;",
	"&ordf;", "&laquo;",
	"&not;", "&shy;",
	"&reg;", "&macr;",
	"&deg;", "&plusmn;",
	"&sup2",
	"&sup3;", "&acute;",
	"&micro;", "&para;",
	"&middot;", "&cedil;",
	"&sup1;", "&ordm;",
	"&raquo;", "&frac14;",
	"&frac12;", "&frac34;",
	"&iquest;", "&times;",
	"&divide;", "&ETH;",
	"&eth;", "&THORN;",
	"&thorn;", "&AElig;",
	"&aelig;", "&OElig;",
	"&oelig;", "&Aring;",
	"&Oslash;", "&Ccedil;",
	"&ccedil;", "&szlig;",
	"&Ntilde;", "&ntilde;",
]

class Html:
    @staticmethod
    def checkRe(f):
        @wrap(f)
        def inner(obj):
            try:
                f(obj)
            except:
                raise AttributeError('re.error')
            else:
                return f(obj)
        return inner


    @staticmethod
    def get_end_digits(string):
        return re.search(r'\.(\d*)\.html',string).group(1)

    @staticmethod
    def toId(text):
            return re.search('/mixtapes/([\w\/]*)', text).group(1)

    @classmethod
    def find_song_names(cls,text):
        songs = re.findall(r'"title"\:"(.*\w*)",\s?"artist"',text)
        songs = list(cls.remove_ampersands(songs))
        return songs
    
    @classmethod
    def get_duration_from(cls,text):
        return  re.findall(r'"duration"\>(.*\d*)\<',text)

    @staticmethod
    def remove_ampersands(string):
        results = []
        if not isinstance(string,(list,tuple)):
            string = [string]
        for x in string:
            for amps in ampersands:
                x = re.sub(amps,'',x)
            results.append(x)
        return results
                

    def find_name_of_mp3(text):
        songs = re.findall(r'fix.concat\(\s\'(.*\w*)\'',text)
        if songs:
            return [re.sub(' ', '%20',song) for song in songs]




import os
import sys 
import re
import threading
import tempfile
import platform 
import collections
import math
import subprocess
from functools import wraps

class Dict():
    def __init__(self):
        self.data = collection.defaultdict(dict)

    def setAttr(self,name,_list,start=1):
        """Convert list object to a enumerate dictionary"""
        for count,data in enumerate(_list,start=start):
            self.data[count][name] = data
        


def file_size(buf_size):
    """Convert file size and returns user readable size""" 
    if buf_size == 0:
        return '0B'
    
    size_name = ['B','KB','MB','GB','TB','PB','EB','ZB','YB']
    change = int(math.floor(math.log(buf_size,1024)))
    power = math.pow(1024,change)
    result = round(buf_size / power,2)
    return '%s %s'%(result,size_name[change])
 

class Tmp(object):
    """
    Handle all temporary files created by Media player.

    It removes the hanging temparary files that are not destroyed.
    Since we are creating temporary files to store mp3 content, we need 
    to remove these files manually after being used.
    python3 module 'signal' sometimes fail to catch  signal 
    on certain platform like Windows IDLE.
    We create our own, in case user close window or program dies before 
    being able to remove these files. 
    """

    def __init__(self):
        pass

    @staticmethod
    def create():
        """Create a temporary file with suffix __datpiff"""
        return tempfile.NamedTemporaryFile(suffix='_datpiff',delete=False)
         

    @staticmethod
    def removeTmpOnstart():
        """remove all temporary file created by Datpiff on start up"""
        # every tmpfile using media player will have this suffix
        suffix = '_datpiff'
        tmp_dir = tempfile.gettempdir()
        if os.path.isdir(tmp_dir):
            for lf in os.listdir(tmp_dir):
                if suffix in lf:
                    try:
                        lf = '/'.join((tmp_dir,lf))
                        os.remove(lf)
                    except Exception as e:
                        pass


    def threader(f):
        @wraps(f)
        def inner(self,*a,**kw):
            t = threading.Thread(target=f,args=(self,a))
            t.start()
            return t
        return inner



class Path():
    @staticmethod
    def is_dir(path):
        path = path or ''
        return os.path.isdir(path)

    @classmethod
    def join(cls,path='',to=''):
        if not cls.is_dir(path):
            path = os.getcwd()
        return os.path.join(path,to)

    @staticmethod
    def toStandard(name):
        return re.sub('[^A-Za-z0-9_\-\.] ', '', name)




import os 
import sys 
import re
import subprocess
import platform 
import tempfile 
import logging 


def fixdate():
    date = None
    _f = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    if platform.system() == "Linux":
        try:
            _f = 'date +"%m/%d/%y  %I:%M:%S %p"'
            date = subprocess.check_output(_f).decode('utf-8').strip()
            date += ' %(name)s - %(levelname)s - %(message)s'
        except:
            pass
    else:
        date = _f
    return date



class Logger(object):
    logit = logging.getLogger(__name__)
    logging.basicConfig(level=logging.WARNING,format = '%(message)s')
    logfile = os.path.join(tempfile.gettempdir(),'pydatpiff.log')
    handler = logging.FileHandler(logfile)
    handler.setLevel(logging.WARNING)
    _format = logging.Formatter(fixdate(),datefmt="%m/%d/%Y %I:%M:%S %p %Z")
    handler.setFormatter(_format)
    logit.addHandler(handler)


    @classmethod
    def _parseLog(cls,*msg,level='info'):

        msg = ' '.join(msg)
        if level == 'info':
            cls.logit.info(msg)
        elif level == 'warning':
            cls.logit.warning(msg)
        elif level == 'critical':
            cls.logit.critical(msg)


    @classmethod
    def display(cls,*msg):
        cls._parseLog(*msg,level='info')


    @classmethod
    def warn(cls,*msg):
        cls._parseLog(*msg,level='warning')
    

    @classmethod
    def failed(cls,*msg):
        cls._parseLog(*msg,level='critical')
 


class Error(Exception):
    __error__ = {
                 1:'invalid mixtapes object',
                 2:'no mixtapes found'
                }

    code = None
    message='Unknown'


    def __init__(self,code,detail=''):
        self._code = code
        code = self.create(code,detail)
        super().__init__(code)


    def logError(self, error,critical=False):
        if error and error in self.__error__:
            if not critical:
                Logger.warn(self.__error__[error])
            else:
                Logger.failed(self.__error__[error])
                sys.exit(1)


    def show(self,code):
        return self.__error__.get(code)



    @staticmethod
    def makeError(code):
        #code = code or self.message
        cont = []
        for x  in code.split(' '):
            cont.append(x.capitalize())
        cont = ''.join(cont)
        return cont


    @classmethod
    def create(cls,export_to,long_msg=''):
        if isinstance(export_to,str):
            max_e = max(cls.__error__)
            cls.__error__[max_e+1] = Error.makeError(export_to)
            return Error.create(max_e+1)
        elif export_to in cls.__error__:
            for code,error in cls.__error__.items():
                if not isinstance(error,str):
                    return error
                name = Error.makeError(error)
                e = type(name,(cls,),{
                    'code': code,
                    'message':error,
                    })
                cls.__error__[code] = e  
            return "".join((str(cls.__error__[export_to]),'\n'+long_msg))
   


class MixtapesError(Error):
    """ handle all the Media errors"""
    __error__ = {
                 1: 'No Mixtapes Found',
                 2: 'Invalid category selected',
                }


class MediaError(Error):
    """ handle all the Media errors"""
    __error__ = {
                 1: 'no mixtapes found',
                 2: 'invalid mixtapes object',
                 3: 'media album not set',
                 4: 'invaild track index',
                 5: 'song selection error',
                }


class AlbumError(Error):
    __error__ = {
                  1: 'Mixtapes Not Found',
                  2: 'Invalid category selected',
                }


class Mp3Error(Error):
    pass


class PlayerError(Error):
    __error__  = {
                   1: 'vlc not installed',
                 }    


class RequestError(Error):
    # _type = self.__class__._qualname__
    __error__ = {
                 1: 'Invalid url scheme',
                 2: 'Request failed',
                 3: 'Request Non-200 status code',
                }


class BuildError(Error):
    __error__ = {
                 1, 'user selection',
                }




from __init__ import __version__
from setuptools import setup,find_packages

with open('README.md',encoding='utf-8') as f:
    readme = f.read()

#package_dir={'datpiff':'pydatpiff'},
setup( name='pydatpiff',
       version=__version__,
       description='PyDatpiff Mixtape player',
       url='https://github.com/cbedroid/pydatpiff',
       long_description=readme,
       long_description_content_type='text/markdown',
       author='Cornelius Brooks',
       author_email='cbedroid1614@gmail.com',
       install_requires=open('requirements.txt').readline(),
       license='MIT',
       tests_require=['pytest==5.2.4'],
       classifiers=[
                    'License :: OSI Approved :: MIT License',
                    'Intended Audience :: Developers',
                    'Programming Language :: Python :: 3',
                    'Programming Language :: Python :: 3.3',
                    'Programming Language :: Python :: 3.4',
                    'Programming Language :: Python :: 3.6',
                    'Programming Language :: Python :: 3.7',
                    'Programming Language :: Python :: 3.8',
                    'Operating System :: OS Independent',
                    'Topic :: Multimedia :: Sound/Audio :: Players',

                    ],
       zip_safe=False,
       packages=find_packages(exclude=['tests'])
     )

